from typing import Any, Dict, List, Type, Union

import timm
import torch
import torch.nn.functional as F
from timm.layers import BatchNormAct2d, SelectAdaptivePool2d
from timm.models._efficientnet_blocks import SqueezeExcite
from torch import nn


class SelectAdaptivePool1d(nn.Module):
    def __init__(self, pool_type):
        super().__init__()
        self.pool_type = pool_type

        self.flatten = nn.Flatten(1)

        if not pool_type:
            self.pool = nn.Identity()  # pass through
        elif pool_type == "avg":
            self.pool = nn.AdaptiveAvgPool1d(1)
        elif pool_type == "max":
            self.pool = nn.AdaptiveMaxPool1d(1)
        else:
            # TODO: options
            raise NotImplementedError()

    def forward(self, x):
        x = self.pool(x)
        x = self.flatten(x)
        return x


def get_act_layer(name: Union[Type[nn.Module], str] = "relu"):
    """Activation Layer Factory
    Fetching activation layers by name with this function allows export or torch script friendly
    functions to be returned dynamically based on current config.
    """
    if name is None:
        return None
    if not isinstance(name, str):
        # callable, module, etc
        return name
    if not name:
        return None
    name = name.lower()
    if not (is_exportable() or is_scriptable()):
        if name in _ACT_LAYER_ME:
            return _ACT_LAYER_ME[name]
    return _ACT_LAYER_DEFAULT[name]


def create_act_layer(name: Union[Type[nn.Module], str], inplace=None, **kwargs):
    act_layer = get_act_layer(name)
    if act_layer is None:
        return None
    if inplace is None:
        return act_layer(**kwargs)
    try:
        return act_layer(inplace=inplace, **kwargs)
    except TypeError:
        # recover if act layer doesn't have inplace arg
        return act_layer(**kwargs)


def _create_act(act_layer, act_kwargs=None, inplace=False, apply_act=True):
    act_kwargs = act_kwargs or {}
    act_kwargs.setdefault("inplace", inplace)
    act = None
    if apply_act:
        act = create_act_layer(act_layer, **act_kwargs)
    return nn.Identity() if act is None else act


class BatchNormAct1d(nn.BatchNorm1d):
    """BatchNorm + Activation

    This module performs BatchNorm + Activation in a manner that will remain backwards
    compatible with weights trained with separate bn, act. This is why we inherit from BN
    instead of composing it as a .bn member.
    """

    def __init__(
        self,
        num_features,
        eps=1e-5,
        momentum=0.1,
        affine=True,
        track_running_stats=True,
        apply_act=True,
        act_layer=nn.ReLU,
        act_kwargs=None,
        inplace=True,
        drop_layer=None,
    ):
        super().__init__(
            num_features,
            eps=eps,
            momentum=momentum,
            affine=affine,
            track_running_stats=track_running_stats,
        )
        self.drop = drop_layer() if drop_layer is not None else nn.Identity()
        self.act = _create_act(
            act_layer, act_kwargs=act_kwargs, inplace=inplace, apply_act=apply_act
        )

    def forward(self, x):
        # exponential_average_factor is set to self.momentum
        # (when it is available) only so that it gets updated
        # in ONNX graph when this node is exported to ONNX.
        if self.momentum is None:
            exponential_average_factor = 0.0
        else:
            exponential_average_factor = self.momentum

        if self.training and self.track_running_stats:
            # TODO: if statement only here to tell the jit to skip emitting this when it is None
            if self.num_batches_tracked is not None:  # type: ignore[has-type]
                self.num_batches_tracked.add_(1)  # type: ignore[has-type]
                if self.momentum is None:  # use cumulative moving average
                    exponential_average_factor = 1.0 / float(self.num_batches_tracked)
                else:  # use exponential moving average
                    exponential_average_factor = self.momentum

        r"""
        Decide whether the mini-batch stats should be used for normalization rather than the buffers.
        Mini-batch stats are used in training mode, and in eval mode when buffers are None.
        """
        if self.training:
            bn_training = True
        else:
            bn_training = (self.running_mean is None) and (self.running_var is None)

        r"""
        Buffers are only updated if they are to be tracked and we are in training mode. Thus they only need to be
        passed when the update should occur (i.e. in training mode when they are tracked), or when buffer stats are
        used for normalization (i.e. in eval mode when buffers are not None).
        """
        x = F.batch_norm(
            x,
            # If buffers are not to be tracked, ensure that they won't be updated
            (
                self.running_mean
                if not self.training or self.track_running_stats
                else None
            ),
            self.running_var if not self.training or self.track_running_stats else None,
            self.weight,
            self.bias,
            bn_training,
            exponential_average_factor,
            self.eps,
        )
        x = self.drop(x)
        x = self.act(x)
        return x


def forward_se_fixed(self, x):
    x_se = x.mean((2), keepdim=True)
    x_se = self.conv_reduce(x_se)
    x_se = self.act1(x_se)
    x_se = self.conv_expand(x_se)
    return x * self.gate(x_se)


conv2d_to_conv1d: Dict[str, Any] = {
    "in_class": nn.Conv2d,
    "out_class": nn.Conv1d,
    "kwargs_to_copy": [
        "in_channels",
        "out_channels",
        "kernel_size",
        "stride",
        "padding",
        "dilation",
        "groups",
        "bias",
        "padding_mode",
    ],
}

maxpool2d_to_maxpool1d: Dict[str, Any] = {
    "in_class": nn.MaxPool2d,
    "out_class": nn.MaxPool1d,
    "kwargs_to_copy": [
        "kernel_size",
        "stride",
        "padding",
        "dilation",
        "return_indices",
    ],
}

batchnorm2d_to_batchnorm1d: Dict[str, Any] = {
    "in_class": nn.BatchNorm2d,
    "out_class": nn.BatchNorm1d,
    "kwargs_to_copy": [
        "num_features",
        "eps",
        "momentum",
        "affine",
        "track_running_stats",
    ],
}

batchnormact2d_to_batchnormact1d: Dict[str, Any] = {
    "in_class": BatchNormAct2d,
    "out_class": BatchNormAct1d,
    "kwargs_to_copy": [
        "num_features",
        "eps",
        "momentum",
        "affine",
        "track_running_stats",
    ],
}


avgpool2d_to_avgpool1d: Dict[str, Any] = {
    "in_class": SelectAdaptivePool2d,
    "out_class": SelectAdaptivePool1d,
    "kwargs_to_copy": [
        "pool_type",
    ],
}

resnet_2d_to_1d_mapping: List[Dict[str, Any]] = [
    conv2d_to_conv1d,
    maxpool2d_to_maxpool1d,
    batchnorm2d_to_batchnorm1d,
    avgpool2d_to_avgpool1d,
]

efficientnet_2d_to_1d_mapping: List[Dict[str, Any]] = [
    conv2d_to_conv1d,
    maxpool2d_to_maxpool1d,
    batchnormact2d_to_batchnormact1d,
    avgpool2d_to_avgpool1d,
]

all_mapping = resnet_2d_to_1d_mapping + efficientnet_2d_to_1d_mapping


def param2d_to_1d(param):
    if isinstance(param, nn.Parameter):
        return False
    elif isinstance(param, tuple) or isinstance(param, list):
        return param[0]
    else:
        return param


def replace_layers_inplace_recursive(
    module: nn.Module, mapping_config: List[Dict[str, Any]], verbose: bool = True
) -> None:
    # пройдём по всем подмодулям текущего модуля:
    for n, submodule in module.named_children():
        # если у подмодуля есть свои под-подмодули, то пробежимся по всем:
        if len(list(submodule.children())) > 0:
            replace_layers_inplace_recursive(submodule, mapping_config, verbose)

        # если мы дошли до одиночного модуля, осуществим замену
        for conf_item in mapping_config:
            if isinstance(submodule, SqueezeExcite):
                bound_method = forward_se_fixed.__get__(submodule, submodule.__class__)
                setattr(submodule, "forward", bound_method)

            if isinstance(submodule, conf_item["in_class"]):
                # соберём старые значения параметров слоя
                layer_params = {}
                for arg in conf_item["kwargs_to_copy"]:
                    layer_params[arg] = param2d_to_1d(getattr(submodule, arg))

                new_layer = conf_item["out_class"](**layer_params)
                setattr(module, n, new_layer)
                if verbose:
                    print(
                        f'\n{n}: \n {conf_item["in_class"]} -> {conf_item["out_class"]}\n'
                    )


def create_1d_model(
    model_name: str,
    num_classes: int,
    in_chans: int = 1,
    verbose: bool = False,
) -> Type[nn.Module]:
    """
    Create a 1D neural network model using timm framework with layer replacements.

    Creates a model from timm library, replaces specific layers according to a predefined 2d->1d mapping,
    and adapts it for 1D operations. The model is initialized without pretrained weights.

    Args:
        model_name: Name of the timm model to instantiate (e.g., 'resnet18')
        num_classes: Number of output classes for the final classification layer
        in_chans: Number of input channels/features (default: 1)
        verbose: Whether to print model modification details (default: False)

    Returns:
        nn.Module: Modified neural network model adapted for 1D input

    Note:
        - Uses `replace_layers_inplace_recursive` to modify original 2D convolutional layers to 1D equivalents
        - Requires predefined `all_mapping` dictionary specifying layer replacement rules
        - Model created with `pretrained=False` by default

    Example:
        >>> model = create_1d_model('resnet18', num_classes=42, in_chans=1)
        >>> print(model)
    """
    model = timm.create_model(
        model_name=model_name,
        pretrained=False,
        num_classes=num_classes,
        in_chans=in_chans,
    )

    replace_layers_inplace_recursive(model, all_mapping, verbose)

    return model
